from microprediction.live.xraytickers import get_xray_tickers
from microprediction.live.iex import iex_latest_prices, iex_common_stock_with_balance_sheet_tickers
import numpy as np
import math
from getjson import getjson
from microprediction.whereami import TOP
import os
import json
import math


SIZE_TILT_COEF = 0.1
RANK_TILT_COEF = 0.1
LOG_TILT_COEF = 0.1
TILE_NDX = [-2,-1,0,1,2]  # Must be odd
RANK_NDX = TILE_NDX
LOG_NDX = TILE_NDX
NUM_SIZE_PORTFOLIOS = len(TILE_NDX)
NUM_RANK_PORTFOLIOS = len(RANK_NDX)
NUM_LOG_PORTFOLIOS = len(LOG_NDX)
NUM_PORTFOLIOS = NUM_SIZE_PORTFOLIOS + NUM_RANK_PORTFOLIOS
NDX_CAP = (NUM_RANK_PORTFOLIOS-1)/2                       # The index of the cap weight portfolio
NDX_ONE_OVER_N = (NUM_RANK_PORTFOLIOS-1)/2 + NUM_SIZE_PORTFOLIOS + NUM_LOG_PORTFOLIOS     # The index of the 1/n portfolio

XRAY_PORTFOLIO_NAMES = ['xray_' + str(i)+'.json' for i in range(NUM_PORTFOLIOS)]
XRAY_COMMON_JSON = os.path.join(TOP, 'live', 'xrayportfolios.json')

from microprediction.live.xraytickers import STOCK_THRESHOLD


def normalize(w):
    return [ wi/sum(w) for wi in w]


def create_xray_portfolios():
    # One-off determination of portfolio weights, as perturbations of pseudo-index
    from microprediction.live.iexcredentials import get_iex_key
    IEX_KEY = get_iex_key()
    tickers = get_xray_tickers()
    num_tickers = len(tickers)
    prices = iex_latest_prices(tickers=tickers, api_key=IEX_KEY)
    common = iex_common_stock_with_balance_sheet_tickers(api_key=IEX_KEY, tickers=tickers, return_tickers=False, threshold=0.005*STOCK_THRESHOLD)

    common = [ max(0.005, c) for c in common ]

    norm_common = normalize(common)
    norm_cap = normalize([price * cm for price, cm in zip(prices, norm_common)])
    centered_cap = [ (nc-np.mean(norm_cap))/np.std(norm_cap) for nc in norm_cap ]

    norm_log = normalize([ math.log(price * cm) for price, cm in zip(prices, norm_common)])
    centered_log = [(nc - np.mean(norm_log)) / np.std(norm_log) for nc in norm_log]

    norm_rank = normalize([i for i in range(num_tickers)])
    centered_rank = [ (nr - np.mean(norm_rank))/np.std(norm_rank) for nr in norm_rank ]

    xray_portfolios = list()
    for ndx in TILE_NDX:
        cap_tilt = normalize([math.exp(nc * SIZE_TILT_COEF * ndx) for nc in centered_cap])
        xray_portfolios.append(cap_tilt)

    for ndx in LOG_NDX:
        log_tilt = normalize([math.exp(lc * LOG_TILT_COEF * ndx) for lc in centered_log])
        xray_portfolios.append(log_tilt)

    for ndx in RANK_NDX:
        rank_tilt = normalize([math.exp(nr * RANK_TILT_COEF * ndx) for nr in centered_rank])
        xray_portfolios.append(rank_tilt)

    xray_portfolios_dict = dict([(k, p) for k, p in enumerate(xray_portfolios)])

    with open(XRAY_COMMON_JSON, 'wt') as fp:
        json.dump(obj=xray_portfolios_dict, fp=fp)


def get_xray_portfolios():
    """ Retrieve xray portfolios """
    data = getjson('https://raw.githubusercontent.com/microprediction/microprediction/master/microprediction/live/xrayportfolios.json')
    return [ data[str(i)] for i in range(len(data)) ]


if __name__=='__main__':
    CREAT_PORTFOLIOS = False
    if CREAT_PORTFOLIOS:
        # DON'T DO THIS ... SUPPOSED TO BE ONE-OFF
        create_xray_portfolios()
    else:
        portfolios = get_xray_portfolios()
        print(np.shape(portfolios))
        # Check they are all there
        tickers = get_xray_tickers()
        n_tickers = len(tickers)
        assert n_tickers == np.shape(portfolios)[1]





# IEX stock prices
# You can get your api_key at https://iexcloud.io/
# See https://iexcloud.io/docs/api/#quote

from getjson import getjson
from pprint import pprint
import os
import json

DEFAULT_EXCHANGES = ['XNYS', 'BATS', 'ARCX', 'XASE',  'XNAS']  # XPOR


def iex_exchanges(api_key:str):
    url = 'https://cloud.iexapis.com/beta/ref-data/symbols?token=' + api_key
    data = getjson(url)
    return list(set([d['exchange'] for d in data]))


def iex_common_stock_tickers(api_key:str, exchanges=None) -> [str]:
    """ Listing of all common stock tickers """
    if exchanges is None:
        exchanges = DEFAULT_EXCHANGES
    url = 'https://cloud.iexapis.com/beta/ref-data/symbols?token='+api_key
    data = getjson(url)
    return sorted( [ d['symbol'].lower() for d in data if (d['exchange'] in exchanges) and (d['type']=='cs')])


def iex_latest_prices(tickers, api_key:str)->[float]:
    """
        Grab multiple stock prices
        IEX allows only 100 at a time
    """
    SENTINEL = 666
    def price_or_sentinel(s):
        if s is None:
            return SENTINEL
        else:
            if s.get('price') is None:
                return SENTINEL
            else:
                return s.get('price')

    if len(tickers)>100:
        tickerlists = [tickers[i:i + 100] for i in range(0, len(tickers), 100)]
        mids = [ iex_latest_prices(tickers=ticks, api_key=api_key) for ticks in tickerlists ]
        mids_flat = [item for sublist in mids for item in sublist]
        return mids_flat
    else:
        substituted = [ 'meta' if t=='fb' else t for t in tickers]
        symbols = ','.join(substituted)
        paid_url = 'https://cloud.iexapis.com/v1/stock/market/batch?types=price&symbols='+symbols+'&token=' + api_key
        print(paid_url)
        data = getjson(paid_url)
        mids = [ price_or_sentinel(data.get(ticker.upper())) for ticker in tickers]
        return mids


def iex_common_stock_with_balance_sheet_tickers(api_key:str, return_tickers=True, tickers=None, threshold=3000000)->[float]:
    """
        return_tickers: bool   If True, returns list of tickers. If False, returns list of common stock counts
    """
    if tickers is None:
        tickers = iex_common_stock_tickers(api_key=api_key)

    prices = iex_latest_prices(tickers=tickers,api_key=api_key)

    import time
    common = list()
    missing = list()
    small = list()
    for ticker_no, (ticker, price) in enumerate(zip(tickers,prices)):
        paid_url = 'https://cloud.iexapis.com/v1/stock/'+ticker+'/balance-sheet?token=' + api_key
        data = getjson(paid_url, paid_url)
        try:
            cm = data['balancesheet'][0]['commonStock']
            value = float(cm)*float(price)
            if value>threshold:
                if return_tickers:
                    common.append(ticker)
                else:
                    common.append(cm)
            else:
                small.append(ticker)
        except:
            if return_tickers:
                missing.append(ticker)
            else:
                missing.append(ticker)
                common.append(0)

        time.sleep(0.1)
        if ticker_no % 50 == 0:
            print(ticker_no)
            print('There are '+str(len(missing))+' stocks missing balance sheet information ')
            print('There are ' + str(len(common)) + ' stocks with balance sheet that are not too small ')
            print('There are ' + str(len(small)) + ' stocks considered too small ')

    return common


if __name__=='__main__':
    from microprediction.config_private import IEX_KEY
    tickers = ["aapl","abbv","amzn","bac"]
    print(iex_common_stock_with_balance_sheet_tickers(tickers=['googl'], api_key=IEX_KEY))



from getjson import getjson

def pga_contenders():
    return getjson('https://micropredictionmiscstreams.pythonanywhere.com/contenders/names')
  
  
def contender_index(player_name, contenders=None):
    if contenders is None:
      contenders = pga_contenders()
    matches = [ ndx for ndx,name in contenders.items() if name==player_name ]
    if matches:
        return matches[0]
    else:
        return -1


def contender_stream_name(contender_ndx, hole):
    return 'pga_contender_'+str(contender_ndx)+'_hole_'+str(hole)+'.json'



import numpy as np
from microconventions import sign_changes, is_process


# --------------------------------------------------------------------------
#           Miscellaneous utilities
# --------------------------------------------------------------------------

def inv_cdf_walk(inv_cdf, k: int, x0: float) -> float:
    """ Generate sample from sum of independent increments generated by inverse cdf function
         inv_cdf  get_iex_realtime_price   taking (0,1) --> R
         k        int    Number of steps
         x0       float  Starting value
    """
    # See crawler_examples/decastyle_cat for usage example
    x = x0
    for _ in range(k):
        u = np.random.rand()
        dx = inv_cdf(u)
        x = x + dx
    return x


def k_differences(lagged_values,k=1):
    chrono_values = [0.]*k + list(reversed(lagged_values))
    k_diff = list()
    for j in range(k):
        k_diff.extend(np.diff(chrono_values[j:-1:k]))
    return k_diff


def k_diff_std(lagged_values,k=1):
    return np.nanstd(k_differences(lagged_values=lagged_values,k=k))


def k_std(lagged_values,k=1):
    if is_process(lagged_values):
        return k_diff_std(lagged_values,k=k)
    else:
        return np.nanstd(lagged_values)



import json
import os
from microprediction.live.xraytickers import XRAY_TICKERS_JSON
from microprediction.whereami import TOP
from getjson import getjson

XRAY_TICKERS_OLD_JSON = os.path.join(TOP,'live', 'xraytickers_old.json')
XRAY_TICKERS_DISCARDED_JSON = os.path.join(TOP,'live', 'xraytickers_discarded.json')


def get_xray_tickers_discarded() -> [str]:
    """ Retrieve discarded xray tickers """
    url = 'https://raw.githubusercontent.com/microprediction/microprediction/master/microprediction/live/xraytickers_discarded.json'
    data = getjson(url, failover_url=url)
    return [data[str(i)] for i in range(len(data))]


if __name__=='__main__':

    with open(XRAY_TICKERS_OLD_JSON) as fh:
        old_tickers = json.load(fh)

    with open(XRAY_TICKERS_JSON) as fh:
        new_tickers = json.load(fh)

    discarded = [(k, v) for (k,v) in old_tickers.items() if k not in new_tickers]

    with open(XRAY_TICKERS_DISCARDED_JSON, 'wt') as fp:
        json.dump(obj=discarded, fp=fp)



from scipy import stats
import numpy as np
from microprediction.univariate.runningmoments import RunningKurtosis
from microprediction.univariate.distmachine import DistMachine
from microprediction.samplers import evenly_spaced_percentiles


# Moment based approximate skew normal distribution machine based on this idea:
# https://stackoverflow.com/questions/49801071/how-can-i-use-skewnorm-to-produce-a-distribution-with-the-specified-skew
# A bit of a whim so use at your own risk


class SkewDist(DistMachine):

    def __init__(self, state:RunningKurtosis=None, num_interp=500, **ignore):
        state = state or RunningKurtosis()
        super().__init__(state=state,params=None)
        self.cached_samples = None
        self.num_interp = num_interp
        self.percentiles = evenly_spaced_percentiles(num=self.num_interp)

    def update(self, value=None, dt=None, **kwargs):
        self.state.update(value=value, dt=dt)
        self.cached_samples = None

    def inv_cdf(self, p):
        if self.cached_samples is None:
            self.cached_samples = sorted(self.skewed_sample(mean=self.state.mean, sd=self.state.std(),
                                                            skew=self.state.skewness(), num=self.num_interp))
        return np.interp(p, self.percentiles, self.cached_samples)

    @staticmethod
    def skewed_sample(mean, sd, skew, num):
        """
              :returns a collection of samples with roughly the supplied mean, standard deviation and skew
        """
        # see https://gist.github.com/microprediction/2f7b5f062c1267d5baab92aefd3bf0f1 for illustration of fit

        # calculate the degrees of freedom 1 required to obtain the specific
        # skewness statistic, derived from simulations
        loglog_slope = -2.211897875506251
        loglog_intercept = 1.002555437670879
        df2 = 500
        df1 = 10 ** (loglog_slope * np.log10(abs(skew)) + loglog_intercept)

        # sample from F distribution
        fsample = np.sort(stats.f(df1, df2).rvs(size=num))

        # adjust the variance by scaling the distance from each point to the
        # distribution mean by a constant, derived from simulations
        k1_slope = 0.5670830069364579
        k1_intercept = -0.09239985798819927
        k2_slope = 0.5823114978219056
        k2_intercept = -0.11748300123471256

        scaling_slope = abs(skew) * k1_slope + k1_intercept
        scaling_intercept = abs(skew) * k2_slope + k2_intercept

        scale_factor = (sd - scaling_intercept) / scaling_slope
        new_dist = (fsample - np.mean(fsample)) * scale_factor + fsample

        # flip the distribution if specified skew is negative
        if skew < 0:
            new_dist = np.mean(new_dist) - new_dist

        # adjust the distribution mean to the specified value
        samples = new_dist + (mean - np.mean(new_dist))
        return samples


if __name__ == '__main__':
    try:
        import seaborn as sns
        import matplotlib.pyplot as plt
    except ImportError:
        raise (
            'You need to pip install seaborn and matplotlib for this example. Or ')

    '''EXAMPLE'''
    desired_mean = 497.68
    desired_skew = -1.75
    desired_sd = 77.24

    final_dist = SkewDist.skewed_sample(mean=desired_mean, sd=desired_sd, skew=desired_skew, num=225)

    fig, ax = plt.subplots(figsize=(12, 7))
    sns.distplot(final_dist, hist=True, ax=ax, color='green', label='generated distribution')
    sns.distplot(np.random.choice(final_dist, size=225), hist=True, ax=ax, color='red', hist_kws={'alpha': .2},
                 label='sample n=100')
    ax.legend()

    print('Input mean: ', desired_mean)
    print('Result mean: ', np.mean(final_dist), '\n')

    print('Input SD: ', desired_sd)
    print('Result SD: ', np.std(final_dist), '\n')

    print('Input skew: ', desired_skew)
    print('Result skew: ', stats.skew(final_dist))



from microprediction import MicroCrawler
import time, datetime, os
import numpy as np
from pprint import pprint


# A simple testing crawler that takes a pass_callback and a fail_callback.
# Video tutorials are available at https://www.microprediction.com/python-1 to help you
# get started. There's a video explanation of FitCrawler, SequentialCrawler and friends
# at https://www.microprediction.com/fitcrawler


# Remark: You can use this very sparingly :)
# requests.post(url='https://alerts.microprediction.org/slack',json={'text':'hello again via requests'})


class ReportingCrawler(MicroCrawler):

        def __init__(self, write_key, pass_callback, fail_callback, **kwargs):
            """
              :param pass_callback(dict)->bool  function that will be called if things look okay
              :param fail_callback(dict)->bool  function that will be called if things don't
            """
            super().__init__(stop_loss=2, min_lags=0, sleep_time=1, write_key=write_key, quietude=10, verbose=False,**kwargs)
            self.pass_callback = pass_callback
            self.fail_callback = fail_callback
            self.initial_balance = self.get_balance()
            self.max_balance = self.initial_balance
            self.min_balance = self.initial_balance

        def setup(self,**kwargs):
            """ Return error messages if any """
            pass

        def downtime(self,seconds,**ignored):
            super().downtime(seconds=seconds,**ignored)
            self.min_balance = min(self.min_balance, self.get_balance())
            self.max_balance = max(self.max_balance, self.get_balance())

        def teardown(self,**kwargs):
            """ Return error messages if any """
            pass

        def candidate_streams(self,**ignore):
            """ Quickly quickly """
            return [name for name, sponsor in self.get_sponsors().items() if not '~' in name]

        def candidate_delays(self,**ignore):
            """ Fail fast """
            return [70]

        def sample(self, lagged_values, lagged_times=None, delay=None, name=None, **ignore):
            if len(lagged_values or []) > 5:
                return super().sample(lagged_values=lagged_values, lagged_times=lagged_times, delay=delay, name=name)
            else:
                return sorted(
                    np.random.randn(self.num_predictions))

        def default_teardown_errors(self):
            """ Runs a list of ex-post tests """
            # Stale transactions?
            seconds_since_transaction = self.get_elapsed_since_transaction()
            if seconds_since_transaction is None:
                return {'error': 'inactivity', 'message': 'no transactions found'}
            if seconds_since_transaction > 1200:
                return {'error': 'inactivity', 'message': 'more than twenty mins since last transaction',
                        'seconds_since_transaction': seconds_since_transaction}
            # Stale confirms?
            seconds_since_confirm = self.get_elapsed_since_confirm()
            if seconds_since_confirm is None:
                return {'error':'inactivity','message':'no confirmations found'}
            if seconds_since_confirm>1200:
                return {'error':'inactivity','message':'more than twenty mins since last confirm','seconds_since_confirm':seconds_since_confirm}
            # Warn but don't error if balance is the same
            if abs(self.max_balance - self.min_balance) < 1e-5:
                pprint( {'warning': 'inactivity', 'warning message': 'min balance is same as max balance'} )
                print(' ',flush=True)


        def default_setup_errors(self):
            """ Run a bunch of very basic system checks """

            # Are APIs returning?
            try:
                # ... test nullary getters
                for method in ['get_confirms','get_sponsors','get_budgets','get_sponsors',
                               'get_balance','get_errors','get_elapsed_since_transaction',
                               'get_elapsed_since_confirm','maybe_create_key','get_home']:
                    getattr(self,method)()

                method = 'get_cdf'
                self.get_cdf(name='die.json',delay=self.DELAYS[0], values=[0,5,10])
                self.get_cdf(name='tour_11.json',delay=self.DELAYS[-1], values=[-2,-1,0,1,2], as_discrete=True)

                method = 'get_lagged_times'
                self.get_lagged_times('c5_bitcoin.json')
                method = 'get_lagged_values'
                self.get_lagged_values('c5_bitcoin.json')

            except Exception as e:
                return {'error':'setup error','method':method,'message':str(e)}

        def run_and_report(self, timeout=180, name='reporting_crawler'):
            """ Returns error report in form of dict """
            report = {'crawler': name, 'timeout': timeout, 'start_time': time.time(),
                      'start_datetime': str(datetime.datetime.now()),
                      'summary_page':'https://api.microprediction.org/home/'+self.write_key,
                      'base_url':self.base_url,
                      'virtual_env':os.getenv('VIRTUAL_ENV')}
            print(self.write_key)
            pprint(report)

            # Initial checks
            setup_errors = self.default_setup_errors()
            if setup_errors is None:
                try:
                    setup_errors = self.setup()
                except Exception as e:
                    setup_errors = {'error':'error thrown by setup','message':str(e)}

            if setup_errors is None:
                # Run checks
                try:
                    self.run(timeout=timeout)
                    run_errors = None
                except Exception as e:
                    print(str(e), flush=True)
                    passed = False
                    run_errors = {'error':'run error','message':str(e)}
            else:
                run_errors = None

            if setup_errors is None and run_errors is None:
                # Teardown checks
                teardown_errors = self.default_teardown_errors()
                if teardown_errors is None:
                    try:
                        teardown_errors = self.teardown()
                    except Exception as e:
                        teardown_errors = {'error': 'error thrown by teardown', 'message': str(e)}
            else:
                teardown_errors = None

            any_errors = setup_errors or run_errors or teardown_errors
            if setup_errors:
                report.update(setup_errors)
            if run_errors:
                report.update(run_errors)
            if teardown_errors:
                report.update(teardown_errors)

            if not(any_errors) and self.pass_callback is not None:
                successfully_reported_pass = self.pass_callback(report)
                if successfully_reported_pass==False:
                    report.update({'reporting_failure':True})
                    self.fail_callback(report)
            if any_errors and (self.fail_callback is not None):
                self.fail_callback(report)
            pprint(report)


if __name__=="__main__":
    try:
        from microprediction.config_private import pass_callback, fail_callback, FLAMMABLE_COD, FLASHY_COYOTE
    except ImportError as e:
        pass_callback=None
        fail_callback=None

    crawler = ReportingCrawler(base_url='https://devapi.microprediction.org',
                               write_key=FLASHY_COYOTE,
                               pass_callback=pass_callback,
                               fail_callback=fail_callback)
    print(crawler.base_url)
    crawler.run_and_report(timeout=300, name='local test')



from scipy.stats import exponnorm
from microprediction.univariate.fitdist import FitDist
from abc import ABC
from collections import OrderedDict
import math
import numpy as np
from microconventions import evenly_spaced_percentiles, nudged
from copy import deepcopy
from microprediction.univariate.runningmoments import RunningVariance

# Example of a distribution that can be fitted using hyperopt

DEFAULT_EXPNORM_PARAMS = {'g1': 0.5, 'g2': 5.0, 'logK': -2., 'loc': 0.0, 'logScale': 0.0}
DEFAULT_EXPNORM_LOWER = {'g1': 0.001, 'g2': 0.001, 'logK': -5, 'loc': -0.15, 'logScale': -4}
DEFAULT_EXPNORM_UPPER = {'g1': 1.0, 'g2': 15.0, 'logK': 1, 'loc': 0.15, 'logScale': 4.0}
DEFAULT_EXPNORM_HYPER = {'lower_bounds': deepcopy(DEFAULT_EXPNORM_LOWER),
                         'upper_bounds': deepcopy(DEFAULT_EXPNORM_UPPER),
                         'space': None, 'algo': None, 'max_evals': 11}


def dict_sans_none(d):
    """ Return copy of dict without None fields """
    return dict([(k, v) for k, v in d.items() if v is not None])


class ExpNormDist(FitDist, ABC):

    # This moves an anchor point using a two-parameter gain function
    # It judges likelihood based on a symmetric combination of expnorm distributions

    def __init__(self, params: OrderedDict = None, state=None, hyper_params: dict = None):
        self_params = OrderedDict(deepcopy(DEFAULT_EXPNORM_PARAMS))
        if params is not None:
            self_params.update(dict_sans_none(params))
        self_hyper_params = deepcopy(DEFAULT_EXPNORM_HYPER)
        if hyper_params is not None:
            self_hyper_params.update(dict_sans_none(hyper_params))
        self_state = state or {'anchor': None}
        super().__init__(params=self_params, state=self_state, hyper_params=self_hyper_params)
        self.cached_params = None
        self.num_interp = 500
        self.cached_samples = None

    def update(self, value=None, dt=None, **kwargs):
        """ Move the anchor """
        if value is not None:
            dy = value - self.state['anchor'] if self.state['anchor'] is not None else 0.0
            move = self.params['g2'] * math.tanh(self.params['g1'] * dy / self.params['g2'])
            if self.state['anchor'] is not None:
                self.state['anchor'] = self.state['anchor'] + move
            else:
                self.state['anchor'] = value

    def log_likelihood(self, value: float) -> float:
        logK, loc, logScale = self.params['logK'], self.params['loc'], self.params['logScale']
        K = math.exp(logK)
        scale = math.exp(logScale)
        if self.state['anchor'] is None:
            return 0.0
        else:
            x = value - self.state['anchor']
            return math.log(
                exponnorm.pdf(x, K=K, loc=loc, scale=scale) + exponnorm.pdf(-x, K=K, loc=loc, scale=scale) + 0.001)

    def inv_cdf(self, p: float) -> float:
        """ PPF function for mixture of two exponorm distributions """
        if self.cached_samples is None or self.cached_params is None or self.params != self.cached_params or any(
                np.isnan(self.cached_samples)):
            logK, loc, logScale, num = self.params['logK'], self.params['loc'], self.params['logScale'], self.num_interp
            K = math.exp(logK)
            scale = math.exp(logScale)
            percentiles = evenly_spaced_percentiles(num=int(num / 2))
            cdf_1 = [exponnorm.ppf(p, K=K, loc=loc, scale=scale) for p in percentiles]
            cdf_2 = [-exponnorm.ppf(p, K=K, loc=loc, scale=scale) for p in percentiles]
            if len(cdf_1) + len(cdf_2) < num:
                cdf_1 = cdf_1 + [0]
            self.cached_samples = sorted(nudged(cdf_1 + cdf_2))
            self.cached_params = deepcopy(self.params)
        combined_percentiles = evenly_spaced_percentiles(self.num_interp)
        return np.interp(p, combined_percentiles, self.cached_samples)





def get_iex_key():
    try:
        from microprediction.config_private import IEX_KEY
        return IEX_KEY
    except ImportError:
        try:
            from credentials import IEX_KEY
            return IEX_KEY
        except ImportError:
            raise EnvironmentError('The IEX KEY needs to be supplied somehow')


if __name__=='__main__':
    print(get_iex_key())


import os
import json
from getjson import getjson
from microprediction.whereami import TOP
from collections import OrderedDict

RDPS_DESCRIPTIONS = OrderedDict({'e':'energy',
                                 'c':'communication',
                                 'y':'discretionary',
                                 'p':'discretionary',
                                 'f':'financials',
                                 'v':'health',
                                 'i':'industrials',
                                 'b':'materials',
                                 're':'real estate',
                                 'k':'technology',
                                 'u':'utilities'})

# 2022-12-07
RDPS_CAPS_DICT = {'energy':1638445299830,
             'materials':900434969540,
             'industrials':2824657615645,
             'discretionary':3377812181150,
             'staples':2358952321990,
             'health':5196971888210,
             'financials':3788386237860,
             'technology':8626912590550,
             'communication':2439153232630,
             'utilities':1015883727160,
             'real estate':891892003410}

# 2022-12-07
RDPS_PRICES_DICT = {'energy':85.12,
             'materials':82.10,
             'industrials':99.30,
             'discretionary':138.70,
             'staples':76.07,
             'health':138.52,
             'financials':34.59,
             'technology':129.60,
             'communication':49.40,
             'utilities':70.74,
             'real estate':38.10}


RDPS_TICKERS = ['xl'+suffix for suffix in RDPS_DESCRIPTIONS.keys() ]
RDPS_RATIOS = [ RDPS_CAPS_DICT[desc]/RDPS_PRICES_DICT[desc] for e,desc in RDPS_DESCRIPTIONS.items() ]
RDPS_GENERIC_NAMES = [ 'rdps_' + ticker + '.json' for ticker in RDPS_TICKERS ]
RDPS_TICKERS_JSON = os.path.join(TOP,'live', 'rdpstickers.json')
RDPS_TICKERS_REVERSE_JSON = os.path.join(TOP ,'live' ,'rdpstickersreverse.json')


def save_rdps_tickers(tickers):
    tickers_dict = dict([(k, ticker) for k, ticker in enumerate(tickers)])
    with open(RDPS_TICKERS_JSON, 'wt') as fp:
        json.dump(obj=tickers_dict, fp=fp)
    tickers_reverse_dict = dict(sorted([(ticker, k) for k, ticker in enumerate(tickers)]))
    with open(RDPS_TICKERS_REVERSE_JSON, 'wt') as fp:
        json.dump(obj=tickers_reverse_dict, fp=fp)


def get_rdps_tickers() -> [str]:
    """ Retrieve xray tickers """
    url = 'https://raw.githubusercontent.com/microprediction/microprediction/master/microprediction/live/rdpstickers.json'
    data = getjson(url, failover_url=url)
    return [data[str(i)] for i in range(len(data))]


def rdps_index_fraction(prices):
    """ Return percentage of index by cap, given a set of current prices for the sector ETFs """
    assert len(prices)==len(RDPS_RATIOS)
    raw = [ p*r for p,r in zip(prices,RDPS_RATIOS)]
    return [ w/sum(raw) for w in raw]


if __name__=='__main__':
    pass
    # save_rdps_tickers(RDPS_TICKERS)



import math
from microprediction.live.iex import iex_latest_prices, iex_common_stock_with_balance_sheet_tickers
import os
import json
from getjson import getjson
from microprediction.whereami import TOP
from microprediction.live.iex import iex_common_stock_with_balance_sheet_tickers


XRAY_TICKERS_JSON = os.path.join(TOP,'live', 'xraytickers.json')
XRAY_TICKERS_REVERSE_JSON = os.path.join(TOP ,'live' ,'xraytickersreverse.json')

STOCK_THRESHOLD = 250*1000*1000*1000

HARDWIRED_TICKERS=['aapl','abbv','abt','acn','adbe',
                   'aig','amd','amgn','amt','amzn',
                   'avgo','axp','ba','bac','bk','blk','bkng',
                   'bmy','brk-b','c','cat','chtr','cl',
                   'cmcsa','cof','cop','cost','crm','csco',
                   'cvs','dhr','dis','dow','duk','emr',
                   'exc','f','fdx','gd','ge',
                   'gild','gm','goog','googl','gs',
                   'hd','hon','ibm','intc','jnj','jpm',
                   'khc','ko','lin','lly','lmt','low',
                   'ma','mcd','mdlz','mdt','met','meta',
                   'mmm','mo','mrk','ms','msft','nee',
                   'nflx','nke','nvda','orcl','pep',
                   'pfe','pg','pm','pypl','qcom','rtx',
                   'sbux','schw','so','spg','t','tgt',
                   'tmo','tmus','tsla','txn','unh',
                   'unp','ups','usb','v','vz','wba',
                   'wfc','wmt','xom']


def reasonable_threshold():
    # Get common stock value for VSAT, the smallest company in Russell 1000
    from microprediction.live.iexcredentials import get_iex_key
    tickers = ['vsat']
    common = iex_common_stock_with_balance_sheet_tickers(api_key=get_iex_key(), return_tickers=False, tickers=tickers)
    print(common)
    return common


def create_xray_tickers_json():
    """
       Create JSON file storing "official" list of tickers using in the x-ray experiment
    """
    from microprediction.live.iexcredentials import get_iex_key
    IEX_KEY = get_iex_key()
    tickers = iex_common_stock_with_balance_sheet_tickers(api_key=IEX_KEY, return_tickers=True, threshold=STOCK_THRESHOLD)
    save_tickers(tickers=tickers )


def save_tickers(tickers):
    tickers_dict = dict([(k, ticker) for k, ticker in enumerate(tickers)])
    with open(XRAY_TICKERS_JSON, 'wt') as fp:
        json.dump(obj=tickers_dict, fp=fp)
    tickers_reverse_dict = dict(sorted([(ticker, k) for k, ticker in enumerate(tickers)]))
    with open(XRAY_TICKERS_REVERSE_JSON, 'wt') as fp:
        json.dump(obj=tickers_reverse_dict, fp=fp)


def cull_xray_tickers_json(new_threshold):
    """
       Create JSON file storing "official" list of tickers using in the x-ray experiment
    """
    from microprediction.live.iexcredentials import get_iex_key
    IEX_KEY = get_iex_key()
    all_tickers = HARDWIRED_TICKERS
    tickers = iex_common_stock_with_balance_sheet_tickers(api_key=IEX_KEY, tickers=all_tickers, return_tickers=True,threshold=new_threshold)
    save_tickers(tickers=tickers)


def save_hardwired_tickers():
    save_tickers(tickers=HARDWIRED_TICKERS)


def get_xray_tickers() -> [str]:
    """ Retrieve xray tickers """
    url = 'https://raw.githubusercontent.com/microprediction/microprediction/master/microprediction/live/xraytickers.json'
    data = getjson(url, failover_url=url)
    return [data[str(i)] for i in range(len(data))]


def get_yarx_generic_names():
    return [ 'yarx_'+ticker.replace('.','-')+'.json' for ticker in get_xray_tickers() ]


def get_quick_yarx_names():
    return [name.replace('yarx','quick_yarx') for name in get_yarx_generic_names()]


def get_middling_yarx_names():
    return [ 'middling_yarx_'+ticker+'.json' for ticker in get_xray_tickers() ]


def get_slow_yarx_stream_names():
    return [ 'slow_yarx_'+ticker+'.json' for ticker in get_xray_tickers() ]


def save_culled():
    CULL_TICKERS = False
    if CULL_TICKERS:
        from microprediction.live.iexcredentials import get_iex_key
        IEX_KEY = get_iex_key()
        cull_xray_tickers_json()
    else:
        print(reasonable_threshold())
        from microprediction.live.iexcredentials import get_iex_key
        IEX_KEY = get_iex_key()
        cull_xray_tickers_json(new_threshold=STOCK_THRESHOLD * 1)


if __name__ == '__main__':
    tickers = get_xray_tickers()
    print('There are '+str(len(tickers))+' on github JSON ')
    print('There are ' + str(len(HARDWIRED_TICKERS)) + ' hardwired tickers ')
    if False:
        save_hardwired_tickers()



# Examples of scenario generation models
import random, bisect
import numpy as np
import math
from scipy.stats import mode
from microprediction.univariate.cdfvalues import evenly_spaced_percentiles
from microprediction.univariate.processes import is_process
from microconventions.stats_conventions import StatsConventions
from microprediction.univariate.arrivals import approx_mode, approx_mode  # Imported for backward compatibility
from microprediction.univariate.processes import inv_cdf_walk  # Imported for backward compatibility
from scipy.interpolate import interp1d
from collections import Counter
from microprediction.univariate.arrivals import approx_dt


# Rudimentary utilities for generating samples for submission.


# --------------------------------------------------------------------------
#            Empirical discretization and basic utilities
# --------------------------------------------------------------------------

def implied_lattice(lagged_values):
    """ Collection of values which a time series is likely to take, assuming it is not continuous """
    q_lagged_values = np.round(lagged_values, decimals=5)
    q_lagged_lattice = sorted(list(set(q_lagged_values)))
    if 0.75 * len(q_lagged_values) > len(q_lagged_lattice) >= 2:
        dx = mode(np.diff(q_lagged_lattice)).mode[0]
        num_filled = int((q_lagged_lattice[-1] - q_lagged_lattice[0]) / dx) + 1
        filled_lattice = np.linspace(q_lagged_lattice[0], q_lagged_lattice[-1], num=num_filled)
        num_included = np.sum([x in filled_lattice for x in q_lagged_lattice])
        if num_included > 0.9 * len(q_lagged_lattice):
            return filled_lattice
        else:
            return q_lagged_values
    else:
        return q_lagged_values


def project_on_lagged_lattice(values, lagged_values):
    """ Ensure predictions lie on lattice implied by lagged_values
          values  samples that would be submitted were the data continuous
          returns: [ float ] samples that lie on lattice indicated by lagged_values
    """
    lattice = implied_lattice(lagged_values)
    nn_interpolator = interp1d(lattice, lattice, kind='nearest', bounds_error=False, fill_value='extrapolate')
    nearest = nn_interpolator(values)
    return nearest


def center_values_list(values, new_mean=None):
    if (new_mean is not None):
        prior_mean = np.nanmean(values)
        values = [v + new_mean - prior_mean for v in values]
    return values


def scale_values_list(values, new_std=None):
    if (new_std is not None) and (new_std > 0):
        prior_std = np.nanstd(values)
        prior_mean = np.nanmean(values)
        values = [(v - prior_mean) * (new_std / prior_std) + prior_mean for v in values]
    return values


def center_and_scale_values_list(values, new_mean=None, new_std=None):
    """
       Shift mean and/or adjust std
    """
    return scale_values_list(center_values_list(values=values, new_mean=new_mean), new_std=new_std)

# --------------------------------------------------------------------------
#          Recency bootstrappy things
# --------------------------------------------------------------------------


def exponential_bootstrap_projected(lagged_values, decay, num, as_process=None, new_mean=None, new_std=None):
    """ Returns bootstrampped samples taking values only on lattice implied by history """
    values = exponential_bootstrap(lagged=lagged_values, decay=decay, num=num, as_process=as_process)
    values = center_and_scale_values_list(values=values, new_mean=new_mean, new_std=new_std)
    return project_on_lagged_lattice(values=values, lagged_values=lagged_values)


def exponential_bootstrap(lagged, decay, num, as_process=None):
    as_process = as_process or is_process(lagged)
    return differenced_bootstrap(lagged=lagged, decay=decay, num=num) if as_process else independent_bootstrap(
        lagged=lagged, decay=decay, num=num)


def independent_bootstrap(lagged, decay, num):
    """ One parameter jiggled bootstrap favouring more recent observations
          lagged  [ float ]     List most recent observation first
          decay    float        Coefficient in exp(-a k) that weights samples
          num      int          Number of scenarios requested
           :returns  [ float ]  Statistical sample
    """
    weights = list(np.exp([-decay * k for k in range(len(lagged))]))
    empirical_sample = _weighted_random_sample(population=lagged, weights=weights, num=num)
    return StatsConventions.nudged(empirical_sample)


def differenced_bootstrap(lagged, decay, num):
    """ One parameter jiggled bootstrap favouring more recent observations (applied to differences processes) """
    safe_diff_lagged = np.diff(list(lagged) + [0., 0.])
    diff_samples = independent_bootstrap(lagged=safe_diff_lagged, decay=decay, num=num)
    return [lagged[0] + dx for dx in diff_samples]


# --------------------------------------------------------------------------
#            Gaussian
# --------------------------------------------------------------------------

def normal_sample(prediction_mean: float, prediction_std: float, num: int) -> [float]:
    """ Returns normally distributed samples evenly spaced in probability """
    return [prediction_mean + prediction_std * StatsConventions.norminv(p) for p in evenly_spaced_percentiles(num=num)]


def normal_sample_projected(prediction_mean: float, prediction_std: float, num: int, lagged_values):
    """ Returns normally distributed samples taking values only on lattice implied by history """
    values = normal_sample(prediction_mean=prediction_mean, prediction_std=prediction_std, num=num)
    return project_on_lagged_lattice(values=values, lagged_values=lagged_values)


# --------------------------------------------------------------------------
#           Gaussian implied by lagged values
# --------------------------------------------------------------------------

def gaussian_samples(lagged_values, num, as_process=None):
    as_process = as_process or is_process(lagged_values)
    return diff_gaussian_samples(lagged_values=lagged_values, num=num) if as_process else independent_gaussian_samples(
        lagged_values=lagged_values,
        num=num)


def independent_gaussian_samples(lagged_values, num):
    shrunk_std = np.nanstd(list(lagged_values) + [0.01, -0.01])
    shrunk_mean = np.nanmean(lagged_values + [0.0])
    return normal_sample_projected(prediction_mean=shrunk_mean, prediction_std=shrunk_std, num=num)


def diff_gaussian_samples(lagged_values, num):
    """ Samples from differences """
    safe_diff_lagged = np.diff(list(lagged_values) + [0., 0.])
    diff_samples = independent_gaussian_samples(lagged_values=safe_diff_lagged, num=num)
    return [lagged_values[0] + dx for dx in diff_samples]


# --------------------------------------------------------------------------
#            Ad-hoc
# --------------------------------------------------------------------------


def fox_sample(lagged_values, lagged_times, delay, num, name, as_process=None, new_mean=None, new_std=None):
    " Elementary but not completely woeful sampler, used by Malaxable Fox"
    dt = approx_dt(lagged_times)
    lag = max(10, math.ceil(delay / dt))
    print('lag = ' + str(lag))
    is_proc = as_process or ('~' not in name and StatsConventions.is_process(lagged_values))
    if len(lagged_values) < 250 + lag or not is_proc:
        values = exponential_bootstrap(lagged=lagged_values, decay=0.1, num=num, as_process=as_process)
        ret_values = StatsConventions.nudged(project_on_lagged_lattice(values=values, lagged_values=lagged_values))
    else:
        changes = np.diff(list(reversed(lagged_values)), n=lag)
        counter = dict(Counter(changes))
        d = dict(counter)
        num_total = len(changes)
        d1 = dict([(change, round(175 * change_count / num_total)) for change, change_count in d.items()])
        values = list()
        for change, rounded_count in d1.items():
            values.extend([change] * rounded_count)
        change_spray = list(range(-50, 50))
        values.extend(change_spray)
        change_values = values[:num]
        abs_values = [lagged_values[0] + chg for chg in change_values]
        if not len(abs_values) == num:
            # Too many rounded down ... may not be discrete
            abs_values = exponential_bootstrap(lagged=lagged_values, decay=0.1, num=num, as_process=True)
        ret_values = StatsConventions.nudged(project_on_lagged_lattice(values=abs_values, lagged_values=lagged_values))

    ret_values = center_and_scale_values_list(values=ret_values, new_mean=new_mean, new_std=new_std)

    return ret_values




# --------------------------------------------------------------------------
#            Helpers and backward compat
# --------------------------------------------------------------------------

def _weighted_random_sample(weights, num, population=None):
    try:
        return random.choices(weights=weights, k=num, population=population)
    except AttributeError:
        return _alternative_weighted_random_sample(weights=weights, k=num, population=population)


def _alternative_weighted_random_sample(weights, k, population=None):
    """ For way backward compatibility, can probably toss this """
    wrg = _WeightedRandomGenerator(weights)
    ndx = [wrg() for _ in range(k)]
    return ndx if population is None else [population[k] for k in ndx]


class _WeightedRandomGenerator(object):
    def __init__(self, weights):
        self.totals = []
        running_total = 0

        for w in weights:
            running_total += w
            self.totals.append(running_total)

    def next(self):
        rnd = random.random() * self.totals[-1]
        return bisect.bisect_right(self.totals, rnd)

    def __call__(self):
        return self.next()


if __name__ == '__main__':
    values = [-1, 0, 3, 3, 5, 5, 5, 5, 5, 6, 6, 6, 6, 9, 10, 11, 12]
    print(implied_lattice(values))
    values = [-1, 1, 1, -1, 1, 1]
    print(implied_lattice(values))
    values = np.random.randn(10)
    print(list(zip(values, project_on_lagged_lattice(values=values, lagged_values=[-2, -1, 0, 1, 2]))))



from microconventions.stats_conventions import is_discrete, evenly_spaced_percentiles, cdf_values,\
    is_process, nudged, discrete_cdf, sign_changes, quantize, StatsConventions
import numpy as np

# Utilities to help with CDF's

from microprediction.univariate.arrivals import approx_mode

from microprediction.inclusion.scipyinclusion import using_scipy

if using_scipy:
    def robust_std(x):
        """ A robust measure of std """
        # Somewhat arbitrary!
        # https://en.wikipedia.org/wiki/Robust_measures_of_scale
        from scipy.stats import iqr, median_abs_deviation
        iqr_est = iqr(x)/1.349
        mad_est = 1.4826*median_abs_deviation(x)
        return np.nanmean([iqr_est,mad_est])

else:
    def robust_std(x):
        return np.nanstd(x)




import datetime

def yarx_moving(test_name='quick_yarx_appl.json', failover_name='quick_yarx_amd.json') -> bool :
    """
        A lagged indicator of whether the market is open
    """
    import time
    from microprediction import MicroReader
    mr = MicroReader()
    for name in [test_name,failover_name]:
        try:
            lagged_values, lagged_times = mr.get_lagged_values_and_times(name=name)
            name_moving = (time.time()-lagged_times[0])<15*60
            if name_moving:
                return True
        except:
            pass
    return False 
        


def eastern(as_str=True):
    # Get the current time in Eastern Time
    from pytz import timezone
    eastern_zone = timezone('US/Eastern')
    eastern_time = datetime.datetime.now(eastern_zone)
    return eastern_time.strftime("%Y-%m-%d %H:%M:%S.%f") if as_str else eastern_time


def relative_to_4pm():
    today_4pm = datetime.datetime.combine(datetime.date.today(),datetime.time(hour=16))
    dt = eastern(as_str=False).replace(tzinfo=None) - today_4pm
    return dt


def is_near_4pm_eastern(mins=5, seconds=0):
    return abs(relative_to_4pm().total_seconds())<60*mins+seconds


def seconds_until_next_half_hour():
    current_time = datetime.datetime.now()
    return 30*60-(current_time.minute * 60 + current_time.second) % 1800



if __name__=='__main__':
    print(relative_to_4pm().total_seconds())
    print(is_near_4pm_eastern())











if __name__=='__main__':
    print({'yarx_moving':yarx_moving()})
    print(eastern())
